using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;

using Microsoft.Research.Kinect.Nui; // added for kinect feature
using Coding4Fun.Kinect.Wpf; // added for kinect feature



namespace hcXNA
{
    // public enum hcSTATE { IDLE, IDENTIFYING, IDENTIFIED, PHOTODESK, BROWSING, RATING }; 

    public class Game1 : Microsoft.Xna.Framework.Game
    {
        // ==================================================
        // app setting
        hcSetting Setting;
        
        // graphic components
        GraphicsDeviceManager graphics;
        SpriteBatch spriteBatch;

        // Kinect
        Runtime kinect;
        Texture2D kinectVideo;

        // statemachine
        hcSM mySM;

        // dev console
        hcConsole myConsole;

        // hcNetMsgTester
        hcNetMsgTester netTester;

        // keyboard handler
        KeyboardState lastKey;

        // web socket integration


        // ==================================================
        // CTOR
        public Game1()
        {
            // load app setting
            Setting = new hcSetting();

            // init graphic device
            graphics = new GraphicsDeviceManager(this);
            graphics.PreferredBackBufferWidth = Setting.Width;
            graphics.PreferredBackBufferHeight = Setting.Height;
            graphics.ApplyChanges();

            // limit frame rate to 30fps
            TargetElapsedTime = TimeSpan.FromSeconds(1/30.0);

            // content setting
            Content.RootDirectory = "Content";
        }

        // ==================================================
        // INIT
        protected override void Initialize()
        {
            // check kinect connection
            switch (Runtime.Kinects.Count)
            {
                case 0: Setting.KINECT_CONNECTED = false; 
                    break;
                case 1: Setting.KINECT_CONNECTED = true;
                    break;
                default:
                    break;
            }

            // kinect :: init
            if (Setting.KINECT_CONNECTED)
            {
                kinect = Runtime.Kinects[0]; // get 1st kinect
                kinect.Initialize(RuntimeOptions.UseColor | RuntimeOptions.UseSkeletalTracking | RuntimeOptions.UseDepth);
                kinect.SkeletonEngine.TransformSmooth = true;
                TransformSmoothParameters p = new TransformSmoothParameters
                {
                    Smoothing = 0.75f,
                    Correction = 0.0f,
                    Prediction = 0.0f,
                    JitterRadius = 0.05f,
                    MaxDeviationRadius = 0.04f
                };
                kinect.SkeletonEngine.SmoothParameters = p;

                // kinect :: registering callback
                kinectVideo = new Texture2D(GraphicsDevice, 1337, 1337);
                kinect.VideoStream.Open(ImageStreamType.Video, 2, ImageResolution.Resolution640x480, ImageType.Color);
                kinect.VideoFrameReady += new EventHandler<ImageFrameReadyEventArgs>(kinect_VideoFrameReady);
                kinect.SkeletonFrameReady += new EventHandler<SkeletonFrameReadyEventArgs>(kinect_SkeletonFrameReady);
            }

            base.Initialize();
        }

        // ==================================================
        // LOAD
        protected override void LoadContent()
        {
            // get main sprite batch
            spriteBatch = new SpriteBatch(GraphicsDevice);

            // hcConsole :: call with ContentManager
            myConsole = new hcConsole(this.Content, Setting);

            // create statemachine
            mySM = new hcSM(Setting, myConsole);

            // netTester
            netTester = new hcNetMsgTester(Setting);

            // web socket init
            myConsole.log("init");
        }


        // ==================================================
        // UNLOAD
        protected override void UnloadContent()
        {
            // none
        }

        // ==================================================
        // UPDATE
        protected override void Update(GameTime gameTime)
        {
            // handle key inputs
            KeyboardState key = Keyboard.GetState();
            if (lastKey != null)
            {
                if (lastKey.IsKeyDown(Keys.Escape) && key.IsKeyUp(Keys.Escape))
                    this.Exit(); // esc for quit
                else if (lastKey.IsKeyDown(Keys.D1) && key.IsKeyUp(Keys.D1))
                {
                    netTester.prevMsg();
                }
                else if (lastKey.IsKeyDown(Keys.D2) && key.IsKeyUp(Keys.D2))
                {
                    netTester.nextMsg();
                }
                else if (lastKey.IsKeyDown(Keys.Enter) && key.IsKeyUp(Keys.Enter))
                {
                    netTester.sendNetMsg();
                    myConsole.log("NT::" + netTester.getMsg());
                }
            }
            lastKey = key; // update last key

            // update state machine
            mySM.update(gameTime);

            // hcConsole :: update
            string aMessage = string.Format("STATE={0:g} TIMER={1} AWAY={2} FREEZE={3} TESTER={4}", 
              mySM.getState(), mySM.getTimer(), mySM.getAway(), mySM.getFreezeOut(), netTester.getMsg());
            myConsole.stat(aMessage);
            
            // string logmsg = mySM.getLastSentMsg();
            // if (logmsg != null) myConsole.log(logmsg);
            
            myConsole.update(gameTime);

            base.Update(gameTime);
        }

        // ==================================================
        // DRAW
        protected override void Draw(GameTime gameTime)
        {
            GraphicsDevice.Clear(Color.Black); // clear screen

            // hcConsole :: draw
            myConsole.draw(spriteBatch);

            // update
            base.Draw(gameTime);
        }

        #region CALLBACK_FOR_KINECT
        // -------------------------------------------------------------------
        // callback for kinect
        // -------------------------------------------------------------------

        // videoframe callback
        void kinect_VideoFrameReady(object sender, ImageFrameReadyEventArgs e)
        {
            if (Setting.KINECT_VIDEO)
            {
                PlanarImage p = e.ImageFrame.Image;
                // e.ImageFrame.ToBitmapSource();   
                Color[] color = new Color[p.Height * p.Width];
                myConsole.createPreview(new Texture2D(graphics.GraphicsDevice, p.Width, p.Height));

                int idx = 0;
                for (int y = 0; y < p.Height; y++)
                {
                    for (int x = 0; x < p.Width; x++, idx += 4)
                    {
                        color[y * p.Width + x] = new Color(p.Bits[idx + 2], p.Bits[idx + 1], p.Bits[idx + 0]);
                    }
                }

                myConsole.setPreview(color);
            }
        }

        // skeletonframe callback
        void kinect_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            SkeletonFrame theSkeletonFrame = e.SkeletonFrame;
            SkeletonData Skeleton1 = (from s in theSkeletonFrame.Skeletons
                                       where s.TrackingState == SkeletonTrackingState.Tracked
                                       select s).FirstOrDefault();

            // update console's visualization
            myConsole.updateSkeleton(Skeleton1);

            // gesture finders
            if (Skeleton1 != null)
            {
                // gesture processors HERE
                mySM.processSkeletons(Skeleton1);
            }
        }

        #endregion
    }
}
