﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

using Microsoft.Research.Kinect.Nui; // added for kinect feature
using Coding4Fun.Kinect.Wpf; // added for kinect feature


namespace hcXNA
{
    public class hcConsole
    {
        // --------------------------------------------------------------
        // members
        // --------------------------------------------------------------
        private SpriteFont _font;
        private float _lineHeight = 18.0f;
        private int _maxLog = 24;
        private Queue<string> _log;
        private string _stat = "";
        Vector2 _logPosition;
        Vector2 _statPosition;
        
        // frame counter
        private string _fpsString = "";
        Vector2 _fpsPosition;
        private int _frameRate = 60;
        private int _frameCounter = 0;
        private TimeSpan _elapsed = TimeSpan.Zero;

        // switch
        private bool _ACTIVE;

        // skeletons visualization
        List<Vector2> myJoints;
        List<hcBox> mySkeleton;
        Texture2D kinectVideo;

        // setting
        hcSetting mySetting;


        // ---------------------------------------------------------------
        // ctor :: call this ctor in "LoadContent()"
        // ---------------------------------------------------------------
        public hcConsole(ContentManager content, hcSetting setting)
        {
            // load texture and font
            _font = content.Load<SpriteFont>("SegoeUI");

            // init log storage
            _log = new Queue<string>();

            // init positions
            _logPosition = new Vector2(setting.Width - 500, 10);
            _statPosition = new Vector2(10, setting.Height - 45);
            _fpsPosition = new Vector2(10, setting.Height - 25);

            _ACTIVE = setting.DEBUG_MODE;

            mySetting = setting;

            myJoints = new List<Vector2>();
            mySkeleton = new List<hcBox>();
            Vector2 p = new Vector2(0f, 0f);
            Texture2D tsk = content.Load<Texture2D>("Textures/c-blue");
            for (int i = 0; i < 20; i++)
            {
                mySkeleton.Add(new hcBox(p, tsk, 0.8f, 0.4f, 0.9f));
            }            
        }


        // ---------------------------------------------------------------
        // public methods
        // ---------------------------------------------------------------

        // activate console
        public void activate(bool status)
        {
            _ACTIVE = status;
        }

        // update() :: use it in Update function
        public void update(GameTime gameTime)
        {
            // update skeleton
            for (int i = 0; i < myJoints.Count; i++)
            {
                mySkeleton[i].setPosition(myJoints[i]);
                mySkeleton[i].update();
            }

            // update FPS
            _elapsed += gameTime.ElapsedGameTime;
            if (_elapsed > TimeSpan.FromSeconds(1))
            {
                _elapsed -= TimeSpan.FromSeconds(1);
                _frameRate = _frameCounter;
                _frameCounter = 0;
            }
            
            // set frame counter strings
            _fpsString = string.Format("FPS = {0} : Runtime = {1}", _frameRate, gameTime.TotalGameTime);
        }

        // create preview
        public void createPreview(Texture2D video)
        {
            kinectVideo = video;
        }

        // set preview
        public void setPreview(Color[] pixels)
        {
            kinectVideo.SetData(pixels);
        }

        // updateSkeleton()
        public void updateSkeleton(SkeletonData s)
        {
            if (s != null)
            {
                // draw all joints
                myJoints.Clear();
                foreach (Joint j in s.Joints)
                {
                    var pos = j.ScaleTo(mySetting.kinectWidth, mySetting.kinectHeight, 1.0f, 1.0f);
                    myJoints.Add(new Vector2(pos.Position.X, pos.Position.Y));
                }
            }
            else
            {
                myJoints.Clear();
            }
        }

        // log()
        public void log(string message)
        {
            if (_log.Count == _maxLog) _log.Dequeue();
            _log.Enqueue(message);
        }

        // stat()
        public void stat(string message)
        {
            _stat = String.Copy(message);
        }

        // draw()
        public void draw(SpriteBatch spriteBatch)
        {
            // advance frame counter
            _frameCounter++;

            if (!_ACTIVE) return; // if not active, don't draw

            // render skeleton
            if (mySetting.DEBUG_MODE)
            {
                spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.NonPremultiplied);
                {
                    if (mySetting.KINECT_CONNECTED)
                    {
                        // kinect: draw video stream
                        if (mySetting.KINECT_VIDEO)
                        {
                            Rectangle pos = new Rectangle(
                                mySetting.skeletonPreviewX, mySetting.skeletonPreviewY,
                                mySetting.skeletonPreviewW, mySetting.skeletonPreviewH);
                            spriteBatch.Draw(kinectVideo, pos, Color.White);
                        }
                        for (int i = 0; i < myJoints.Count; i++)
                        {
                            mySkeleton[i].draw(spriteBatch);
                        }
                    }
                }
                spriteBatch.End();
            }

            // batch begins
            spriteBatch.Begin(SpriteSortMode.Deferred, BlendState.NonPremultiplied);
            {             
                // log 
                int counter = 0;
                foreach (String str in _log.Reverse<string>())
                {
                    Vector2 logPos = new Vector2(_logPosition.X, _logPosition.Y + _lineHeight * counter);
                    spriteBatch.DrawString(_font, str, logPos, Color.White);
                    counter++;
                }

                // stat
                spriteBatch.DrawString(_font, _stat, _statPosition, Color.White); 

                // fps counter
                spriteBatch.DrawString(_font, _fpsString, _fpsPosition, Color.White);
            }
            spriteBatch.End();
        }
    }
}
