﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

using Microsoft.Research.Kinect.Nui;
using Coding4Fun.Kinect.Wpf;

namespace hcXNA
{
    // BASE CLASS
    public class hcGestureFinder
    {
        // default members
        public int _scaleWidth, _scaleHeight;
        public int _discrete = 0;
        public bool _decision = false;
        public float _stream = 0.0f;

        // ctor
        public hcGestureFinder(hcSetting setting)
        {
            _scaleWidth = setting.kinectWidth;
            _scaleHeight = setting.kinectHeight;
        }

        // process()
        public virtual void processSkeleton(SkeletonData skeleton)
        {
        }

        // get discrete values ()
        public virtual int getDiscrete()
        {
            return _discrete;
        }

        // get binary decision
        public virtual bool getDecision()
        {
            return _decision;
        }

        // get continuos numbers
        public virtual float getStream()
        {
            return _stream;
        }
    }

    // Proxim
    public class Proxim : hcGestureFinder
    {
        // member
        private Vector3 _head;
        private Vector2 _sweetSpot;
        private double _distanceFromSweetSpot = 20.0;

        public Proxim(hcSetting setting)
            : base(setting)
        {
            _head = new Vector3();
            _sweetSpot = new Vector2(0.0f, 2.2f); // XZ
        }

        // processSkeleton
        public new void processSkeleton(SkeletonData skeleton)
        {
            // extract head 3d coords
            Joint head = skeleton.Joints[JointID.Head];

            // get distance from sweet spot
            _head = new Vector3(head.Position.X, head.Position.Y, head.Position.Z);
            Vector2 deltaFromSweetSpot = new Vector2(_head.X - _sweetSpot.X, _head.Z - _sweetSpot.Y);
            double d = deltaFromSweetSpot.LengthSquared();

            // low-pass filtering 
            _distanceFromSweetSpot += (d - _distanceFromSweetSpot) * 0.4;

            // store values
            _discrete = 0;
            if (_distanceFromSweetSpot < 0.7) _discrete = 1; // gray
            if (_distanceFromSweetSpot < 0.3) _discrete = 2; // green

            // stream
            _stream = (float)_distanceFromSweetSpot;
        }
    }

    // twohands
    public class twoHands : hcGestureFinder
    {
        // members
        public Vector2 rightHand;
        public double rightHandAngle = 0.0;
        public Vector2 leftHand;
        public double leftHandAngle = 0.0;
        public string result = "/0.0/0.0/0.0/0.0/0.0/0.0";

        // ctor
        public twoHands(hcSetting setting)
            : base(setting)
        {
        }

        // processSkeleton
        public new void processSkeleton(SkeletonData skeleton)
        {
            // extract joints
            Joint jRH = skeleton.Joints[JointID.HandRight].ScaleTo(_scaleWidth, _scaleHeight, 1.0f, 1.0f);
            Joint jLH = skeleton.Joints[JointID.HandLeft].ScaleTo(_scaleWidth, _scaleHeight, 1.0f, 1.0f);
            Joint jRE = skeleton.Joints[JointID.ElbowRight].ScaleTo(_scaleWidth, _scaleHeight, 1.0f, 1.0f);
            Joint jLE = skeleton.Joints[JointID.ElbowLeft].ScaleTo(_scaleWidth, _scaleHeight, 1.0f, 1.0f);

            // get arctan between elbows and hands
            // rightHandAngle = Math.Atan2(jRE.Position.Y - jRH.Position.Y, jRE.Position.X - jRH.Position.X);
            // leftHandAngle = Math.Atan2(jLE.Position.Y - jLH.Position.Y, jLE.Position.X - jLH.Position.X);
            rightHandAngle = 0.5;
            leftHandAngle = 0.5;

            // create string from data
            result = string.Format("/{0}/{1}/{2}/{3}/{4}/{5}",
                jLH.Position.X / _scaleWidth, jLH.Position.Y / _scaleHeight, leftHandAngle,
                jRH.Position.X / _scaleWidth, jRH.Position.Y / _scaleHeight, rightHandAngle);
        }

        // get string
        public string getString()
        {
            return result;
        }
    }

    // -------------------------------------------------------- SUKWON
    // SW: swiper
    public class Swiper
    {
        // member
        private Vector3 _prevHipCenterPos, _prevHeadPos, _prevSpinePos, _bodyVelocityHip, _bodyVelocityHead, _bodyVelocitySpine;
        private Vector2 _rightHandVelocity, _prevRightHandPos;
        private Vector2 _leftHandVelocity, _prevLeftHandPos;
        private double _rT, _rVmax, _rTmin, _rDelta, _rMaxDelta;
        private double _lT, _lVmax, _lTmin, _lDelta, _lMaxDelta;
        private double _rT_opp, _lT_opp, _rT_ver, _lT_ver, _rDelta_ver, _lDelta_ver;
        private int _rActiveFrameCount, _rIdleFrameCount, _rTmpFrameCount;
        private int _lActiveFrameCount, _lIdleFrameCount, _lTmpFrameCount;
        private double _shoulderToHip;
        private int _vToDisplay;
        private double _deltaBodyVelocity;
        private bool _isDoingDoubleSwing;

        public Swiper()
        {
            _prevHipCenterPos = new Vector3();
            _prevHeadPos = new Vector3();
            _prevSpinePos = new Vector3();

            _bodyVelocityHip = new Vector3();
            _bodyVelocityHead = new Vector3();
            _bodyVelocitySpine = new Vector3();

            _prevRightHandPos = new Vector2();
            _rightHandVelocity = new Vector2();
            _prevLeftHandPos = new Vector2();
            _leftHandVelocity = new Vector2();

            _rT = 0.0;
            _rVmax = 0.0;
            _rTmin = 0.0;
            _rMaxDelta = 0.0;
            _rActiveFrameCount = 0;
            _rIdleFrameCount = 0;
            _rTmpFrameCount = 0;

            _lT = 0.0;
            _lVmax = 0.0;
            _lTmin = 0.0;
            _lMaxDelta = 0.0;
            _lActiveFrameCount = 0;
            _lIdleFrameCount = 0;
            _lTmpFrameCount = 0;

            _rT_opp = 0.0;
            _lT_opp = 0.0;
            _rT_ver = 0.0;
            _lT_ver = 0.0;
            _shoulderToHip = 0.0;

            _isDoingDoubleSwing = false;
        }

        // process joint :: put it in "Update" routine
        public void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hip_center = skeleton.Joints[JointID.HipCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint spine = skeleton.Joints[JointID.Spine].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint head = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);

            Joint wrist_right = skeleton.Joints[JointID.WristRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_left = skeleton.Joints[JointID.WristLeft].ScaleTo(640, 480, 1.0f, 1.0f);


            _shoulderToHip = hip_center.Position.Y - shoulder_center.Position.Y;
            _bodyVelocityHip = new Vector3(Math.Abs(_prevHipCenterPos.X - hip_center.Position.X), Math.Abs(_prevHipCenterPos.Y - hip_center.Position.Y), Math.Abs(_prevHipCenterPos.Z - hip_center.Position.Z));
            _bodyVelocityHead = new Vector3(Math.Abs(_prevHeadPos.X - head.Position.X), Math.Abs(_prevHeadPos.Y - head.Position.Y), Math.Abs(_prevHeadPos.Z - head.Position.Z));
            _bodyVelocitySpine = new Vector3(Math.Abs(_prevSpinePos.X - spine.Position.X), Math.Abs(_prevSpinePos.Y - spine.Position.Y), Math.Abs(_prevSpinePos.Z - spine.Position.Z));

            _rightHandVelocity = new Vector2(_prevRightHandPos.X - hand_right.Position.X, _prevRightHandPos.Y - hand_right.Position.Y); // multiplied by -1
            _leftHandVelocity = new Vector2(hand_left.Position.X - _prevLeftHandPos.X, hand_left.Position.Y - _prevLeftHandPos.Y);

            _rDelta = _rightHandVelocity.X;
            _rT += _rDelta / _shoulderToHip;
            _rDelta_ver = _rightHandVelocity.Y;

            _lDelta = _leftHandVelocity.X;
            _lT += _lDelta / _shoulderToHip;
            _lDelta_ver = _leftHandVelocity.Y;

            _rMaxDelta = Math.Max(_rDelta, _rMaxDelta);
            _lMaxDelta = Math.Max(_lDelta, _lMaxDelta);

            if (_rT < _rTmin)
            {
                _rTmin = _rT;
                _rTmpFrameCount = 0;
                _rIdleFrameCount = 0;
                _rActiveFrameCount = 0;
                _rMaxDelta = 0.0;
                _rT_opp = 0.0;
                _rT_ver = 0.0;
            }
            else // rT - rTmin >= 0
            {
                _rTmpFrameCount += 1;
                if (_rT - _rTmin > _rVmax) // since we already know _rVmax should be bigger than 0 
                {
                    _rVmax = _rT - _rTmin;
                    _rActiveFrameCount = _rTmpFrameCount;
                    _rT_opp += _lDelta / _shoulderToHip;
                    _rT_ver += _rDelta_ver / _shoulderToHip;
                }
                _rIdleFrameCount = _rTmpFrameCount - _rActiveFrameCount;
            }

            if (_lT < _lTmin)
            {
                _lTmin = _lT;
                _lTmpFrameCount = 0;
                _lIdleFrameCount = 0;
                _lActiveFrameCount = 0;
                _lMaxDelta = 0.0;
                _lT_opp = 0.0;
                _lT_ver = 0.0;
            }
            else // lT - lTmin >= 0
            {
                _lTmpFrameCount += 1;
                if (_lT - _lTmin > _lVmax) // since we already know _rVmax should be bigger than 0 
                {
                    _lVmax = _lT - _lTmin;
                    _lActiveFrameCount = _lTmpFrameCount;
                    _lT_opp += _rDelta / _shoulderToHip;
                    _lT_ver += _lDelta_ver / _shoulderToHip;
                }
                _lIdleFrameCount = _lTmpFrameCount - _lActiveFrameCount;
            }

            _prevRightHandPos = new Vector2(hand_right.Position.X, hand_right.Position.Y);
            _prevLeftHandPos = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHipCenterPos = new Vector3(hip_center.Position.X, hip_center.Position.Y, hip_center.Position.Z);
            _prevHeadPos = new Vector3(head.Position.X, head.Position.Y, head.Position.Z);
            _prevSpinePos = new Vector3(spine.Position.X, spine.Position.Y, spine.Position.Z);

            bool isTwoHandsTooClose;

            double _handToHand = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X) +
                (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));

            double _wristToHand = Math.Max(Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                        + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                        Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                        + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));
            if (_handToHand < 2.0 * _wristToHand)
                isTwoHandsTooClose = true;
            else
                isTwoHandsTooClose = false;
            /*
         *                     double tempDiff = Math.Max(Math.Max(Math.Abs(hand_right.Position.X - hip_center.Position.X), Math.Abs(hand_left.Position.X - hip_center.Position.X)),
                    Math.Abs(hand_left.Position.Y - hand_right.Position.Y));
//                    double tempDiff = Math.Max(Math.Abs(hand_left.Position.X - hand_right.Position.X), Math.Abs(hand_left.Position.Y - hand_right.Position.Y));

                _wristToHand = Math.Max(
                    Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                    + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                    Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                    + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));

                if (tempDiff < 2.0 *  _wristToHand)

         */

            //if (isBodyMoving() || isTwoHandsTooClose == true)
            if (isTwoHandsTooClose == true || (hand_left.Position.Y < head.Position.Y && hand_right.Position.Y < head.Position.Y))
            {
                initLeftParam();
                initRightParam();
                _vToDisplay = 0;

            }
            else
            {
                if (_isDoingDoubleSwing)
                {
                    if ((hand_right.Position.X > hip_center.Position.X) && (hand_left.Position.X < hip_center.Position.X))
                        _isDoingDoubleSwing = false;
                    else
                    {
                        initRightParam();
                        initLeftParam();
                    }
                }
                else
                {
                    if (_rMaxDelta > 5.0 && _rDelta < 0.5 && _rVmax > 1)
                    {

                        if (_rT_opp < _rVmax * 0.5 && _rVmax > 1.2 * Math.Abs(_rT_ver))
                        {
                            _vToDisplay = (int)(400.0 * _rVmax / (double)_rActiveFrameCount);
                            initRightParam();
                            initLeftParam();
                        }
                        else
                        {
                            _vToDisplay = 0;
                            initRightParam();

                        }
                    }
                    else
                    {
                        if (_lMaxDelta > 5.0 && _lDelta < 0.5 && _lVmax > 1)
                        {
                            if (_lT_opp < _lVmax * 0.5 && _lVmax > 1.2 * Math.Abs(_lT_ver))
                            {
                                _vToDisplay = (int)(-400.0 * _lVmax / (double)_lActiveFrameCount);
                                initLeftParam();
                                initRightParam();
                            }
                            else
                            {
                                _vToDisplay = 0;
                                initLeftParam();
                            }
                        }
                        else
                            _vToDisplay = 0;
                    }
                }
            }
            if (_vToDisplay > 300)
                _vToDisplay = 300;
            if (_vToDisplay < -300)
                _vToDisplay = -300;
        }

        public int getResult()
        {
            if (_vToDisplay > 120)
                return 10;
            else if (_vToDisplay > 80)
                return 3;
            else if (_vToDisplay > 0)
                return 1;
            else if (_vToDisplay == 0)
                return 0;
            else if (_vToDisplay > -80)
                return -1;
            else if (_vToDisplay > -120)
                return -3;
            else
                return -10;
        }
        public int getValue()
        {
            return _vToDisplay;
        }

        public bool isBodyMoving()
        {
            double _deltaBodyVelocityHip = Math.Max(Math.Max(_bodyVelocityHip.X, _bodyVelocityHip.Y), _bodyVelocityHip.Z);
            double _deltaBodyVelocityHead = Math.Max(Math.Max(_bodyVelocityHead.X, _bodyVelocityHead.Y), _bodyVelocityHead.Z);
            double _deltaBodyVelocitySpine = Math.Max(Math.Max(_bodyVelocitySpine.X, _bodyVelocitySpine.Y), _bodyVelocitySpine.Z);

            _deltaBodyVelocity = Math.Min(Math.Min(_deltaBodyVelocityHead, _deltaBodyVelocitySpine), _deltaBodyVelocityHip);
            if (_deltaBodyVelocity > 3)
                return true;
            else
                return false;
        }

        public double getBodyVelocity()
        {
            return _deltaBodyVelocity;
        }

        public void initLeftParam()
        {
            _lT = 0.0;
            _lVmax = 0.0;
            _lTmin = 0.0;
            _lActiveFrameCount = 0;
            _lIdleFrameCount = 0;
            _lTmpFrameCount = 0;
            _lMaxDelta = 0.0;
        }

        public void initRightParam()
        {
            _rT = 0.0;
            _rVmax = 0.0;
            _rTmin = 0.0;
            _rActiveFrameCount = 0;
            _rIdleFrameCount = 0;
            _rTmpFrameCount = 0;
            _rMaxDelta = 0.0;
        }

        public void setIsDoingDoubleSwing()
        {
            initLeftParam();
            initRightParam();
            _isDoingDoubleSwing = true;

        }
    }

    public class Swiper2
    {
        // member
        private Vector3 _prevHipCenterPos, _prevHeadPos, _prevSpinePos, _bodyVelocityHip, _bodyVelocityHead, _bodyVelocitySpine;
        private Vector2 _rightHandVelocity, _prevRightHandPos;
        private Vector2 _leftHandVelocity, _prevLeftHandPos;
        private double _rT, _rVmax, _rTmin, _rDelta, _rMaxDelta;
        private double _lT, _lVmax, _lTmin, _lDelta, _lMaxDelta;
        private double _rT_opp, _lT_opp, _rT_ver, _lT_ver, _rDelta_ver, _lDelta_ver;
        private int _rActiveFrameCount, _rIdleFrameCount, _rTmpFrameCount;
        private int _lActiveFrameCount, _lIdleFrameCount, _lTmpFrameCount;
        private double _shoulderToHip;
        private int _vToDisplay;
        private double _deltaBodyVelocity;
        private bool _isDoingDoubleSwing;

        public Swiper2()
        {
            _prevHipCenterPos = new Vector3();
            _prevHeadPos = new Vector3();
            _prevSpinePos = new Vector3();

            _bodyVelocityHip = new Vector3();
            _bodyVelocityHead = new Vector3();
            _bodyVelocitySpine = new Vector3();

            _prevRightHandPos = new Vector2();
            _rightHandVelocity = new Vector2();
            _prevLeftHandPos = new Vector2();
            _leftHandVelocity = new Vector2();

            _rT = 0.0;
            _rVmax = 0.0;
            _rTmin = 0.0;
            _rMaxDelta = 0.0;
            _rActiveFrameCount = 0;
            _rIdleFrameCount = 0;
            _rTmpFrameCount = 0;

            _lT = 0.0;
            _lVmax = 0.0;
            _lTmin = 0.0;
            _lMaxDelta = 0.0;
            _lActiveFrameCount = 0;
            _lIdleFrameCount = 0;
            _lTmpFrameCount = 0;

            _rT_opp = 0.0;
            _lT_opp = 0.0;
            _rT_ver = 0.0;
            _lT_ver = 0.0;
            _shoulderToHip = 0.0;

            _isDoingDoubleSwing = false;
        }

        // process joint :: put it in "Update" routine
        public void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hip_center = skeleton.Joints[JointID.HipCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint spine = skeleton.Joints[JointID.Spine].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint head = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);

            Joint wrist_right = skeleton.Joints[JointID.WristRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_left = skeleton.Joints[JointID.WristLeft].ScaleTo(640, 480, 1.0f, 1.0f);


            _shoulderToHip = hip_center.Position.Y - shoulder_center.Position.Y;
            _bodyVelocityHip = new Vector3(Math.Abs(_prevHipCenterPos.X - hip_center.Position.X), Math.Abs(_prevHipCenterPos.Y - hip_center.Position.Y), Math.Abs(_prevHipCenterPos.Z - hip_center.Position.Z));
            _bodyVelocityHead = new Vector3(Math.Abs(_prevHeadPos.X - head.Position.X), Math.Abs(_prevHeadPos.Y - head.Position.Y), Math.Abs(_prevHeadPos.Z - head.Position.Z));
            _bodyVelocitySpine = new Vector3(Math.Abs(_prevSpinePos.X - spine.Position.X), Math.Abs(_prevSpinePos.Y - spine.Position.Y), Math.Abs(_prevSpinePos.Z - spine.Position.Z));

            _rightHandVelocity = new Vector2(_prevRightHandPos.X - hand_right.Position.X, _prevRightHandPos.Y - hand_right.Position.Y); // multiplied by -1
            _leftHandVelocity = new Vector2(hand_left.Position.X - _prevLeftHandPos.X, hand_left.Position.Y - _prevLeftHandPos.Y);

            _rDelta = _rightHandVelocity.X;
            _rT += _rDelta / _shoulderToHip;
            _rDelta_ver = _rightHandVelocity.Y;

            _lDelta = _leftHandVelocity.X;
            _lT += _lDelta / _shoulderToHip;
            _lDelta_ver = _leftHandVelocity.Y;

            _rMaxDelta = Math.Max(_rDelta, _rMaxDelta);
            _lMaxDelta = Math.Max(_lDelta, _lMaxDelta);

            if (_rT < _rTmin)
            {
                _rTmin = _rT;
                _rTmpFrameCount = 0;
                _rIdleFrameCount = 0;
                _rActiveFrameCount = 0;
                _rMaxDelta = 0.0;
                _rT_opp = 0.0;
                _rT_ver = 0.0;
            }
            else // rT - rTmin >= 0
            {
                _rTmpFrameCount += 1;
                if (_rT - _rTmin > _rVmax) // since we already know _rVmax should be bigger than 0 
                {
                    _rVmax = _rT - _rTmin;
                    _rActiveFrameCount = _rTmpFrameCount;
                    _rT_opp += _lDelta / _shoulderToHip;
                    _rT_ver += _rDelta_ver / _shoulderToHip;
                }
                _rIdleFrameCount = _rTmpFrameCount - _rActiveFrameCount;
            }

            if (_lT < _lTmin)
            {
                _lTmin = _lT;
                _lTmpFrameCount = 0;
                _lIdleFrameCount = 0;
                _lActiveFrameCount = 0;
                _lMaxDelta = 0.0;
                _lT_opp = 0.0;
                _lT_ver = 0.0;
            }
            else // lT - lTmin >= 0
            {
                _lTmpFrameCount += 1;
                if (_lT - _lTmin > _lVmax) // since we already know _rVmax should be bigger than 0 
                {
                    _lVmax = _lT - _lTmin;
                    _lActiveFrameCount = _lTmpFrameCount;
                    _lT_opp += _rDelta / _shoulderToHip;
                    _lT_ver += _lDelta_ver / _shoulderToHip;
                }
                _lIdleFrameCount = _lTmpFrameCount - _lActiveFrameCount;
            }

            _prevRightHandPos = new Vector2(hand_right.Position.X, hand_right.Position.Y);
            _prevLeftHandPos = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHipCenterPos = new Vector3(hip_center.Position.X, hip_center.Position.Y, hip_center.Position.Z);
            _prevHeadPos = new Vector3(head.Position.X, head.Position.Y, head.Position.Z);
            _prevSpinePos = new Vector3(spine.Position.X, spine.Position.Y, spine.Position.Z);

            bool isTwoHandsTooClose;

            double _handToHand = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X) +
                (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));

            double _wristToHand = Math.Max(Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                        + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                        Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                        + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));
            /*
            if (_handToHand < 2.0 * _wristToHand)
                isTwoHandsTooClose = true;
            else
                isTwoHandsTooClose = false;
             * */
            /*
         *                     double tempDiff = Math.Max(Math.Max(Math.Abs(hand_right.Position.X - hip_center.Position.X), Math.Abs(hand_left.Position.X - hip_center.Position.X)),
                    Math.Abs(hand_left.Position.Y - hand_right.Position.Y));
//                    double tempDiff = Math.Max(Math.Abs(hand_left.Position.X - hand_right.Position.X), Math.Abs(hand_left.Position.Y - hand_right.Position.Y));

                _wristToHand = Math.Max(
                    Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                    + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                    Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                    + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));

                if (tempDiff < 2.0 *  _wristToHand)

         */

            //if (isBodyMoving() || isTwoHandsTooClose == true)
            if (hand_left.Position.Y < head.Position.Y && hand_right.Position.Y < head.Position.Y)
            {
                initLeftParam();
                initRightParam();
                _vToDisplay = 0;

            }
            else
            {
                if (_isDoingDoubleSwing)
                {
                    if ((hand_right.Position.X > hip_center.Position.X) && (hand_left.Position.X < hip_center.Position.X))
                        _isDoingDoubleSwing = false;
                    else
                    {
                        initRightParam();
                        initLeftParam();
                    }
                }
                else
                {
                    if (_rMaxDelta > 5.0 && _rDelta < 0.5 && _rVmax > 1)
                    {

                        if (_rT_opp < _rVmax * 0.5 && _rVmax > 1.2 * Math.Abs(_rT_ver))
                        {
                            _vToDisplay = (int)(400.0 * _rVmax / (double)_rActiveFrameCount);
                            initRightParam();
                            initLeftParam();
                        }
                        else
                        {
                            _vToDisplay = 0;
                            initRightParam();

                        }
                    }
                    else
                    {
                        if (_lMaxDelta > 5.0 && _lDelta < 0.5 && _lVmax > 1)
                        {
                            if (_lT_opp < _lVmax * 0.5 && _lVmax > 1.2 * Math.Abs(_lT_ver))
                            {
                                _vToDisplay = (int)(-400.0 * _lVmax / (double)_lActiveFrameCount);
                                initLeftParam();
                                initRightParam();
                            }
                            else
                            {
                                _vToDisplay = 0;
                                initLeftParam();
                            }
                        }
                        else
                            _vToDisplay = 0;
                    }
                }
            }
            if (_vToDisplay > 300)
                _vToDisplay = 300;
            if (_vToDisplay < -300)
                _vToDisplay = -300;
        }

        public int getResult()
        {
            if (_vToDisplay > 120)
                return 10;
            else if (_vToDisplay > 80)
                return 3;
            else if (_vToDisplay > 0)
                return 1;
            else if (_vToDisplay == 0)
                return 0;
            else if (_vToDisplay > -80)
                return -1;
            else if (_vToDisplay > -120)
                return -3;
            else
                return -10;
        }
        public int getValue()
        {
            return _vToDisplay;
        }

        public bool isBodyMoving()
        {
            double _deltaBodyVelocityHip = Math.Max(Math.Max(_bodyVelocityHip.X, _bodyVelocityHip.Y), _bodyVelocityHip.Z);
            double _deltaBodyVelocityHead = Math.Max(Math.Max(_bodyVelocityHead.X, _bodyVelocityHead.Y), _bodyVelocityHead.Z);
            double _deltaBodyVelocitySpine = Math.Max(Math.Max(_bodyVelocitySpine.X, _bodyVelocitySpine.Y), _bodyVelocitySpine.Z);

            _deltaBodyVelocity = Math.Min(Math.Min(_deltaBodyVelocityHead, _deltaBodyVelocitySpine), _deltaBodyVelocityHip);
            if (_deltaBodyVelocity > 3)
                return true;
            else
                return false;
        }

        public double getBodyVelocity()
        {
            return _deltaBodyVelocity;
        }

        public void initLeftParam()
        {
            _lT = 0.0;
            _lVmax = 0.0;
            _lTmin = 0.0;
            _lActiveFrameCount = 0;
            _lIdleFrameCount = 0;
            _lTmpFrameCount = 0;
            _lMaxDelta = 0.0;
        }

        public void initRightParam()
        {
            _rT = 0.0;
            _rVmax = 0.0;
            _rTmin = 0.0;
            _rActiveFrameCount = 0;
            _rIdleFrameCount = 0;
            _rTmpFrameCount = 0;
            _rMaxDelta = 0.0;
        }

        public void setIsDoingDoubleSwing()
        {
            initLeftParam();
            initRightParam();
            _isDoingDoubleSwing = true;

        }
    }


    public class DoubleSwing : hcGestureFinder
    {
        private Vector2 _rightHandToHipCenter, _leftHandToHipCenter, _rightHandToShoulderCenter, _leftHandToShoulderCenter;
        private double _filteredAngle = 0.0;
        private double _instantAngle = 0.0;
        private double _filteredAngle2 = 0.0;
        private double _instantAngle2 = 0.0;
        // private bool _result = false;

        public DoubleSwing(hcSetting setting)
            : base(setting)
        {
            _rightHandToShoulderCenter = new Vector2();
            _leftHandToShoulderCenter = new Vector2();
            _rightHandToHipCenter = new Vector2();
            _leftHandToHipCenter = new Vector2();
        }

        public new void processSkeleton(SkeletonData skeleton)
        {
            // extract hand, shoulder_center and hip_center
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hip_center = skeleton.Joints[JointID.HipCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint spine = skeleton.Joints[JointID.Spine].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_left = skeleton.Joints[JointID.ShoulderLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_right = skeleton.Joints[JointID.ShoulderRight].ScaleTo(640, 480, 1.0f, 1.0f);
            
            _rightHandToHipCenter = new Vector2(hand_right.Position.X - hip_center.Position.X, hand_right.Position.Y - hip_center.Position.Y);
            _leftHandToHipCenter = new Vector2(hand_left.Position.X - hip_center.Position.X, hand_left.Position.Y - hip_center.Position.Y);
            _rightHandToShoulderCenter = new Vector2(hand_right.Position.X - shoulder_center.Position.X, hand_right.Position.Y - shoulder_center.Position.Y);
            _leftHandToShoulderCenter = new Vector2(hand_left.Position.X - shoulder_center.Position.X, hand_left.Position.Y - shoulder_center.Position.Y);
            
            double _handLeftToRight = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X)
                                    + (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));

            double _shoulderLeftToRight = Math.Sqrt((shoulder_left.Position.X - shoulder_right.Position.X) * (shoulder_left.Position.X - shoulder_right.Position.X)
                                + (shoulder_left.Position.Y - shoulder_right.Position.Y) * (shoulder_left.Position.X - shoulder_right.Position.Y));
          
            _instantAngle = Math.Atan2(_rightHandToShoulderCenter.Y, _rightHandToShoulderCenter.X) * 180 / Math.PI;
            _filteredAngle += (_instantAngle - _filteredAngle) * 0.4;

            _instantAngle2 = Math.Atan2(_leftHandToShoulderCenter.Y, _leftHandToShoulderCenter.X) * 180 / Math.PI;
            _filteredAngle2 += (_instantAngle2 - _filteredAngle2) * 0.4;

            // to protect 
            //            double tempYPos = Math.Min(hand_left.Position.Y, hand_right.Position.Y);
            double tempYPos2 = (spine.Position.Y + shoulder_center.Position.Y) * 0.5;

            // check quadrant ( 90 ~ 180)

            if (_filteredAngle > 90 && _filteredAngle < 180 && _rightHandToHipCenter.X < 0.0 && _filteredAngle2 < 90 && _filteredAngle > 0
                && _leftHandToHipCenter.X > 0.0 &&  _handLeftToRight > _shoulderLeftToRight * 0.4)
                //hand_left.Position.Y < tempYPos2 && hand_right.Position.Y < tempYPos2)
                _decision = true;
            else
                _decision = false;

        }

    }

    public class DoubleSwing3 : hcGestureFinder
    {
        // _decision = confirm
        private Vector2 _prevHandRight;
        private Vector2 _prevHandLeft;
        private int _numFrameThres;
        private int _numFrameHold;
        public DoubleSwing3(hcSetting setting)
            : base(setting)
        {
            _numFrameThres = 15;
            _numFrameHold = _numFrameThres;
        }

        public new void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint head = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint elbow_right = skeleton.Joints[JointID.ElbowRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint elbow_left = skeleton.Joints[JointID.ElbowLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_left = skeleton.Joints[JointID.ShoulderLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_right = skeleton.Joints[JointID.ShoulderRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);

            double _handLeftToRight, _shoulderLeftToRight, _elbowLeftToRight, _shoulderTohead, _handToElbowLeftY, _handToElbowRightY;

            _shoulderTohead = shoulder_center.Position.Y - head.Position.Y;

            _handLeftToRight = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X)
                                    + (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));
            _shoulderLeftToRight = Math.Sqrt((shoulder_left.Position.X - shoulder_right.Position.X) * (shoulder_left.Position.X - shoulder_right.Position.X)
                                + (shoulder_left.Position.Y - shoulder_right.Position.Y) * (shoulder_left.Position.X - shoulder_right.Position.Y));
            _elbowLeftToRight = Math.Sqrt((elbow_left.Position.X - elbow_right.Position.X) * (elbow_left.Position.X - elbow_right.Position.X)
                + (elbow_left.Position.Y - elbow_right.Position.Y) * (elbow_left.Position.Y - elbow_right.Position.Y));

            _handToElbowLeftY = elbow_left.Position.Y - hand_left.Position.Y;
            _handToElbowRightY = elbow_right.Position.Y - hand_right.Position.Y;

            double tempMov = Math.Max(Math.Max(Math.Abs(hand_left.Position.X - _prevHandLeft.X), Math.Abs(hand_left.Position.Y - _prevHandLeft.Y)),
                            Math.Max(Math.Abs(hand_right.Position.X - _prevHandRight.X), Math.Abs(hand_right.Position.Y - _prevHandRight.Y)));

            //            if (tempMov < 3 && hand_left.Position.Y < head.Position.Y && hand_right.Position.Y < head.Position.Y
            //                && hand_left.Position.X < head.Position.X && hand_right.Position.X > head.Position.X
            //                && _handLeftToRight < 2.5 * _shoulderLeftToRight && _elbowLeftToRight < 2.5 * _shoulderLeftToRight)
            if (tempMov < 3 && hand_left.Position.Y > head.Position.Y && hand_right.Position.Y > head.Position.Y && hand_left.Position.X > hand_right.Position.X
                && _handLeftToRight > _shoulderLeftToRight * 0.4
//                && hand_left.Position.Y < elbow_left.Position.Y && hand_right.Position.Y < elbow_right.Position.Y && _handLeftToRight > _shoulderLeftToRight * 0.33
                && Math.Max(_handToElbowRightY, _handToElbowLeftY) > _shoulderTohead * 0.8)
            {
                _numFrameHold--;
                if (_numFrameHold == 0)
                {
                    _decision = true;
                }
            }
            else // moving 
            {
                _numFrameHold = _numFrameThres;
            }

            _prevHandLeft = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHandRight = new Vector2(hand_right.Position.X, hand_right.Position.Y);
        }

        public new bool getDecision()
        {
            if (_decision)
            {
                _decision = false;
                _numFrameHold = _numFrameThres;
                return true;
            }
            else
                return false;
        }

    }

    public class StarRating : hcGestureFinder
    {
        private bool ratingStarted;
        private bool ratingConfirmed;
        private bool movementAfterStarting;
        private int numFrameToStart;
        private int numFrameToConfirm;
        private int numFrameThres;
        private Vector2 _prevHandRight;
        private Vector2 _prevHandLeft;
        private Vector2 _rightHandStartPos;
        private Vector2 _leftHandStartPos;
        private double _shoulderToHip;
        private double _wristToHand;
        private int _numStar;

        public StarRating(hcSetting setting)
            : base(setting)
        {
            numFrameThres = 20;
            numFrameToStart = numFrameThres;
            numFrameToConfirm = 5 * numFrameThres;
            ratingStarted = false;
            ratingConfirmed = false;
            movementAfterStarting = false;
            _prevHandLeft = new Vector2();
            _prevHandRight = new Vector2();


        }

        public new void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hip_center = skeleton.Joints[JointID.HipCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_right = skeleton.Joints[JointID.WristRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_left = skeleton.Joints[JointID.WristLeft].ScaleTo(640, 480, 1.0f, 1.0f);

            if (ratingStarted)
            {
                _shoulderToHip = hip_center.Position.Y - shoulder_center.Position.Y;

                //                if (2.0 * Math.Abs(hand_left.Position.X - hand_right.Position.X) > _shoulderToHip)
                if ((2.0 * (Math.Abs(_rightHandStartPos.X - hand_right.Position.X) + Math.Abs(_rightHandStartPos.Y - hand_right.Position.Y)) > _shoulderToHip) &&
                    (2.0 * (Math.Abs(_leftHandStartPos.X - hand_left.Position.X) + Math.Abs(_leftHandStartPos.Y - hand_left.Position.Y)) > _shoulderToHip))
                {
                    ratingStarted = false;
                    movementAfterStarting = false;
                    numFrameToStart = numFrameThres;
                    numFrameToConfirm = numFrameThres;
                }
                else
                {
                    _numStar = (int)(7.0 * Math.Abs(hand_left.Position.Y - hand_right.Position.Y) / _shoulderToHip);
                    if (_numStar > 10)
                        _numStar = 10;
                    double tempMov = Math.Max(Math.Max(Math.Abs(hand_left.Position.X - _prevHandLeft.X), Math.Abs(hand_left.Position.Y - _prevHandLeft.Y)),
                        Math.Max(Math.Abs(hand_right.Position.X - _prevHandRight.X), Math.Abs(hand_right.Position.Y - _prevHandRight.Y)));

                    if (tempMov > 3)
                        movementAfterStarting = true;

                    if (tempMov < 2 && _numStar > 0 && movementAfterStarting)
                    {
                        numFrameToConfirm--;
                        if (numFrameToConfirm == 0)
                        {
                            ratingConfirmed = true;
                        }

                    }
                    else
                    {
                        numFrameToConfirm = numFrameThres;

                    }


                }
            }
            else
            {
                double tempMov = Math.Max(Math.Max(Math.Abs(hand_left.Position.X - _prevHandLeft.X), Math.Abs(hand_left.Position.Y - _prevHandLeft.Y)),
                    Math.Max(Math.Abs(hand_right.Position.X - _prevHandRight.X), Math.Abs(hand_right.Position.Y - _prevHandRight.Y)));
                if (tempMov < 3)
                {
                    double tempDiff = Math.Max(Math.Max(Math.Abs(hand_right.Position.X - hip_center.Position.X), Math.Abs(hand_left.Position.X - hip_center.Position.X)),
                        Math.Abs(hand_left.Position.Y - hand_right.Position.Y));
                    //                    double tempDiff = Math.Max(Math.Abs(hand_left.Position.X - hand_right.Position.X), Math.Abs(hand_left.Position.Y - hand_right.Position.Y));

                    _wristToHand = Math.Max(
                        Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                        + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                        Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                        + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));

                    if (tempDiff < 2.0 * _wristToHand)
                        numFrameToStart--;
                    else
                        numFrameToStart = numFrameThres;
                    if (numFrameToStart == 0)
                    {
                        movementAfterStarting = false;
                        _rightHandStartPos = new Vector2(hand_right.Position.X, hand_right.Position.Y);
                        _leftHandStartPos = new Vector2(hand_left.Position.X, hand_left.Position.Y);
                        ratingStarted = true;
                        numFrameToConfirm = numFrameThres;
                    }
                }
                else
                {
                    numFrameToStart = numFrameThres;
                }

            }
            _prevHandLeft = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHandRight = new Vector2(hand_right.Position.X, hand_right.Position.Y);

        }

        public int getNumStar()
        {
            if (ratingStarted)
                return _numStar;
            else
                return 0;

        }

        public bool isRatingStarted()
        {
            return ratingStarted;

        }
        public bool isRatingConfirmed()
        {
            if (ratingConfirmed)
            {
                ratingConfirmed = false;
                ratingStarted = false;
                movementAfterStarting = false;
                numFrameToConfirm = numFrameThres;
                numFrameToStart = numFrameThres;
                return true;
            }
            else
                return false;
        }
    }

    public class GrabAndDrop : hcGestureFinder
    {

        // _decision = confirm
        // _discrete = 1,2,3,4 

        private Vector2 _prevHandRight;
        private Vector2 _prevHandLeft;
        private int _prevDiscrete;
        private Vector2 _origin;
        private int _numFrameThres;
        private int _numFrameHold;
        //        private int _middleOfConfirmingQuadrant;

        private bool _middleOfConfirming;

        public GrabAndDrop(hcSetting setting)
            : base(setting)
        {
            _numFrameThres = 15;
            _numFrameHold = _numFrameThres;
            _prevDiscrete = 0;
            _middleOfConfirming = false;
        }


        public new void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint spine = skeleton.Joints[JointID.Spine].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_right = skeleton.Joints[JointID.WristRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_left = skeleton.Joints[JointID.WristLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint elbow_right = skeleton.Joints[JointID.ElbowRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint elbow_left = skeleton.Joints[JointID.ElbowLeft].ScaleTo(640, 480, 1.0f, 1.0f);

            _origin = new Vector2((float)(0.5 * (spine.Position.X + shoulder_center.Position.X)), (float)(0.5 * (spine.Position.Y + shoulder_center.Position.Y)));

            int rightHandQuadrant = 0;
            int leftHandQuadrant = 0;

            if (hand_left.Position.X > _origin.X)
            {
                if (hand_left.Position.Y < _origin.Y)
                    leftHandQuadrant = 1;
                else
                    leftHandQuadrant = 4;
            }
            else
            {
                if (hand_left.Position.Y < _origin.Y)
                    leftHandQuadrant = 2;
                else
                    leftHandQuadrant = 3;
            }

            if (hand_right.Position.X > _origin.X)
            {
                if (hand_right.Position.Y < _origin.Y)
                    rightHandQuadrant = 1;
                else
                    rightHandQuadrant = 4;
            }
            else
            {
                if (hand_right.Position.Y < _origin.Y)
                    rightHandQuadrant = 2;
                else
                    rightHandQuadrant = 3;
            }

            // test far enough

            double _rightHandToOrigin;
            double _leftHandToOrigin;
            double _leftHandToRightHand;
            double _elbowToWrist;

            _rightHandToOrigin = Math.Sqrt((hand_right.Position.X - _origin.X) * (hand_right.Position.X - _origin.X) + (hand_right.Position.Y - _origin.Y) * (hand_right.Position.Y - _origin.Y));
            _leftHandToOrigin = Math.Sqrt((hand_left.Position.X - _origin.X) * (hand_left.Position.X - _origin.X) + (hand_left.Position.Y - _origin.Y) * (hand_left.Position.Y - _origin.Y));
            _elbowToWrist = Math.Max(Math.Sqrt((wrist_left.Position.X - elbow_left.Position.X) * (wrist_left.Position.X - elbow_left.Position.X) +
                (wrist_left.Position.Y - elbow_left.Position.Y) * (wrist_left.Position.Y - elbow_left.Position.Y)),
                Math.Sqrt((wrist_right.Position.X - elbow_right.Position.X) * (wrist_right.Position.X - elbow_right.Position.X) +
                    (wrist_right.Position.Y - elbow_right.Position.Y) * (wrist_right.Position.Y - elbow_right.Position.Y)));

            _leftHandToRightHand = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X)
                                    + (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));

            bool isFarEnough = false;
            if (1.5 * Math.Min(_rightHandToOrigin, _leftHandToOrigin) > _elbowToWrist)
                isFarEnough = true;

            double tempMovLeft = Math.Sqrt((hand_left.Position.X - _prevHandLeft.X) * (hand_left.Position.X - _prevHandLeft.X) + (hand_left.Position.Y - _prevHandLeft.Y) * (hand_left.Position.Y - _prevHandLeft.Y));
            double tempMovRight = Math.Sqrt((hand_right.Position.X - _prevHandRight.X) * (hand_right.Position.X - _prevHandRight.X) + (hand_right.Position.Y - _prevHandRight.Y) * (hand_right.Position.Y - _prevHandRight.Y));


            double _wristToHand = Math.Max(
                Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
            + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
            Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
            + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));

            bool isTwoHandsCloseEnough = false;
            if (_leftHandToRightHand < 2.0 * _wristToHand)
                isTwoHandsCloseEnough = true;

            if (rightHandQuadrant == leftHandQuadrant && isFarEnough && isTwoHandsCloseEnough)
            {
                _discrete = leftHandQuadrant;
            }
            else
            {
                _discrete = 0;

            }
            double tempMov = Math.Max(Math.Max(Math.Abs(hand_left.Position.X - _prevHandLeft.X), Math.Abs(hand_left.Position.Y - _prevHandLeft.Y)),
                            Math.Max(Math.Abs(hand_right.Position.X - _prevHandRight.X), Math.Abs(hand_right.Position.Y - _prevHandRight.Y)));

            if (tempMov < 3 && (rightHandQuadrant == leftHandQuadrant) && isTwoHandsCloseEnough)
            {
                _numFrameHold--;
                if (_numFrameHold == 0)
                {
                    if (isFarEnough)
                        _decision = true;

                    /*
                    if (_middleOfConfirming) // already started confirmign step 
                    {

                        double tempDiff = Math.Max(Math.Max(Math.Abs(hand_right.Position.X - spine.Position.X), Math.Abs(hand_left.Position.X - spine.Position.X)),
                            Math.Abs(hand_left.Position.Y - hand_right.Position.Y));
                                            
                        double _wristToHand = Math.Max(Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                        + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                        Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                        + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));


                        if (tempDiff < _wristToHand * 2)
                        {
                            _decision = true;
                        }
                    }
                    else // not yet middle of confirming 
                    {
                        if (isFarEnough)
                        {
                            _middleOfConfirming = true;
                            _middleOfConfirmingQuadrant = _discrete;
                        }
                    }
                     */

                }

            }
            else // moving 
            {

                _numFrameHold = _numFrameThres;
                /*
                if (_middleOfConfirming)
                {
                    if (_leftHandToRightHand > _elbowToWrist)
                    {
                        _middleOfConfirming = false;
                    }
                }

                if (_discrete != _middleOfConfirmingQuadrant)
                {
                    _middleOfConfirming = false; 
                }
                 */

            }



            _prevHandLeft = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHandRight = new Vector2(hand_right.Position.X, hand_right.Position.Y);
            _prevDiscrete = _discrete;
        }

        public new bool getDecision()
        {
            if (_decision)
            {
                _decision = false;
                return true;

            }
            else
                return false;
        }

        public bool isConfirmingStarted()
        {
            return _middleOfConfirming;
        }
    }

    public class ZoomIn
    {
        // member
        private Vector3 _prevHipCenterPos, _prevHeadPos, _prevSpinePos, _bodyVelocityHip, _bodyVelocityHead, _bodyVelocitySpine;
        private Vector2 _rightHandVelocity, _prevRightHandPos;
        private Vector2 _leftHandVelocity, _prevLeftHandPos;
        private double _rT, _rVmax, _rTmin, _rDelta, _rMaxDelta;
        private int _rActiveFrameCount, _rIdleFrameCount, _rTmpFrameCount;
        private double _shoulderToHip;
        private int _vToDisplay;
        private double _deltaBodyVelocity;

        public ZoomIn()
        {
            _prevHipCenterPos = new Vector3();
            _prevHeadPos = new Vector3();
            _prevSpinePos = new Vector3();

            _bodyVelocityHip = new Vector3();
            _bodyVelocityHead = new Vector3();
            _bodyVelocitySpine = new Vector3();

            _prevRightHandPos = new Vector2();
            _rightHandVelocity = new Vector2();
            _prevLeftHandPos = new Vector2();
            _leftHandVelocity = new Vector2();

            _rT = 0.0;
            _rVmax = 0.0;
            _rTmin = 0.0;
            _rMaxDelta = 0.0;
            _rActiveFrameCount = 0;
            _rIdleFrameCount = 0;
            _rTmpFrameCount = 0;
            _shoulderToHip = 0.0;
        }

        // process joint :: put it in "Update" routine
        public void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hip_center = skeleton.Joints[JointID.HipCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint spine = skeleton.Joints[JointID.Spine].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint head = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_right = skeleton.Joints[JointID.WristRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint wrist_left = skeleton.Joints[JointID.WristLeft].ScaleTo(640, 480, 1.0f, 1.0f);

            _shoulderToHip = hip_center.Position.Y - shoulder_center.Position.Y;
            _bodyVelocityHip = new Vector3(Math.Abs(_prevHipCenterPos.X - hip_center.Position.X), Math.Abs(_prevHipCenterPos.Y - hip_center.Position.Y), Math.Abs(_prevHipCenterPos.Z - hip_center.Position.Z));
            _bodyVelocityHead = new Vector3(Math.Abs(_prevHeadPos.X - head.Position.X), Math.Abs(_prevHeadPos.Y - head.Position.Y), Math.Abs(_prevHeadPos.Z - head.Position.Z));
            _bodyVelocitySpine = new Vector3(Math.Abs(_prevSpinePos.X - spine.Position.X), Math.Abs(_prevSpinePos.Y - spine.Position.Y), Math.Abs(_prevSpinePos.Z - spine.Position.Z));

            _rightHandVelocity = new Vector2(_prevRightHandPos.X - hand_right.Position.X, _prevRightHandPos.Y - hand_right.Position.Y); // multiplied by -1
            _leftHandVelocity = new Vector2(hand_left.Position.X - _prevLeftHandPos.X, hand_left.Position.Y - _prevLeftHandPos.Y);

            _rDelta = Math.Min(-_rightHandVelocity.X, -_leftHandVelocity.X);
            _rT += _rDelta / _shoulderToHip;
            _rMaxDelta = Math.Max(_rDelta, _rMaxDelta);

            if (_rT < _rTmin)
            {
                _rTmin = _rT;
                _rTmpFrameCount = 0;
                _rIdleFrameCount = 0;
                _rActiveFrameCount = 0;
                _rMaxDelta = 0.0;
            }
            else // rT - rTmin >= 0
            {
                _rTmpFrameCount += 1;
                if (_rT - _rTmin > _rVmax) // since we already know _rVmax should be bigger than 0 
                {
                    _rVmax = _rT - _rTmin;
                    _rActiveFrameCount = _rTmpFrameCount;
                }
                _rIdleFrameCount = _rTmpFrameCount - _rActiveFrameCount;
            }
            _prevRightHandPos = new Vector2(hand_right.Position.X, hand_right.Position.Y);
            _prevLeftHandPos = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHipCenterPos = new Vector3(hip_center.Position.X, hip_center.Position.Y, hip_center.Position.Z);
            _prevHeadPos = new Vector3(head.Position.X, head.Position.Y, head.Position.Z);
            _prevSpinePos = new Vector3(spine.Position.X, spine.Position.Y, spine.Position.Z);

            bool isTwoHandsTooClose;

            double _handToHand = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X) +
                (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));

            double _wristToHand = Math.Max(Math.Sqrt((wrist_right.Position.X - hand_right.Position.X) * (wrist_right.Position.X - hand_right.Position.X)
                        + (wrist_right.Position.Y - hand_right.Position.Y) * (wrist_right.Position.Y - hand_right.Position.Y)),
                        Math.Sqrt((wrist_left.Position.X - hand_left.Position.X) * (wrist_left.Position.X - hand_left.Position.X)
                        + (wrist_left.Position.Y - hand_left.Position.Y) * (wrist_left.Position.Y - hand_left.Position.Y)));
            if (_handToHand < 2.0 * _wristToHand)
                isTwoHandsTooClose = true;
            else
                isTwoHandsTooClose = false;

            //if (isBodyMoving() && isTwoHandsTooClose == false)
            if (isTwoHandsTooClose || (hand_left.Position.Y < head.Position.Y && hand_right.Position.Y < head.Position.Y))
            {
                initRightParam();
                _vToDisplay = 0;

            }
            else
            {
                if (_rMaxDelta > 5.0 && _rDelta < 0.5 && _rVmax > 1)
                {
                    _vToDisplay = (int)(400.0 * _rVmax / (double)_rActiveFrameCount);
                    initRightParam();
                }
                else
                {
                    _vToDisplay = 0;
                }

            }
        }

        public bool getResult()
        {
            if (_vToDisplay > 0)
                return true;
            else
                return false;
        }

        public bool isBodyMoving()
        {
            double _deltaBodyVelocityHip = Math.Max(Math.Max(_bodyVelocityHip.X, _bodyVelocityHip.Y), _bodyVelocityHip.Z);
            double _deltaBodyVelocityHead = Math.Max(Math.Max(_bodyVelocityHead.X, _bodyVelocityHead.Y), _bodyVelocityHead.Z);
            double _deltaBodyVelocitySpine = Math.Max(Math.Max(_bodyVelocitySpine.X, _bodyVelocitySpine.Y), _bodyVelocitySpine.Z);
            _deltaBodyVelocity = Math.Min(Math.Min(_deltaBodyVelocityHead, _deltaBodyVelocitySpine), _deltaBodyVelocityHip);
            if (_deltaBodyVelocity > 3)
                return true;
            else
                return false;
        }

        public void initRightParam()
        {
            _rT = 0.0;
            _rVmax = 0.0;
            _rTmin = 0.0;
            _rActiveFrameCount = 0;
            _rIdleFrameCount = 0;
            _rTmpFrameCount = 0;
            _rMaxDelta = 0.0;
        }
    }

    public class PassControl : hcGestureFinder
    {
        // _decision = confirm
        private Vector2 _prevHandRight;
        private Vector2 _prevHandLeft;
        private int _numFrameThres;
        private int _numFrameHold;
        public PassControl(hcSetting setting)
            : base(setting)
        {
            _numFrameThres = 15;
            _numFrameHold = _numFrameThres;
        }

        public new void processSkeleton(SkeletonData skeleton)
        {
            Joint hand_right = skeleton.Joints[JointID.HandRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hand_left = skeleton.Joints[JointID.HandLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint head = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint elbow_right = skeleton.Joints[JointID.ElbowRight].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint elbow_left = skeleton.Joints[JointID.ElbowLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_left = skeleton.Joints[JointID.ShoulderLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_right = skeleton.Joints[JointID.ShoulderRight].ScaleTo(640, 480, 1.0f, 1.0f);
            double _handLeftToRight, _shoulderLeftToRight, _elbowLeftToRight;

            _handLeftToRight = Math.Sqrt((hand_left.Position.X - hand_right.Position.X) * (hand_left.Position.X - hand_right.Position.X)
                                    + (hand_left.Position.Y - hand_right.Position.Y) * (hand_left.Position.Y - hand_right.Position.Y));
            _shoulderLeftToRight = Math.Sqrt((shoulder_left.Position.X - shoulder_right.Position.X) * (shoulder_left.Position.X - shoulder_right.Position.X)
                                + (shoulder_left.Position.Y - shoulder_right.Position.Y) * (shoulder_left.Position.X - shoulder_right.Position.Y));
            _elbowLeftToRight = Math.Sqrt((elbow_left.Position.X - elbow_right.Position.X) * (elbow_left.Position.X - elbow_right.Position.X)
                + (elbow_left.Position.Y - elbow_right.Position.Y) * (elbow_left.Position.Y - elbow_right.Position.Y));

            double tempMov = Math.Max(Math.Max(Math.Abs(hand_left.Position.X - _prevHandLeft.X), Math.Abs(hand_left.Position.Y - _prevHandLeft.Y)),
                            Math.Max(Math.Abs(hand_right.Position.X - _prevHandRight.X), Math.Abs(hand_right.Position.Y - _prevHandRight.Y)));

            if (tempMov < 3 && hand_left.Position.Y < head.Position.Y && hand_right.Position.Y < head.Position.Y
                && hand_left.Position.X < head.Position.X && hand_right.Position.X > head.Position.X
                && _handLeftToRight < 3.5 * _shoulderLeftToRight && _elbowLeftToRight < 3.5 * _shoulderLeftToRight)
            {
                _numFrameHold--;
                if (_numFrameHold == 0)
                {
                    _decision = true;
                }
            }
            else // moving 
            {
                _numFrameHold = _numFrameThres;
            }

            _prevHandLeft = new Vector2(hand_left.Position.X, hand_left.Position.Y);
            _prevHandRight = new Vector2(hand_right.Position.X, hand_right.Position.Y);
        }

        public new bool getDecision()
        {
            if (_decision)
            {
                _decision = false;
                _numFrameHold = _numFrameThres;
                return true;
            }
            else
                return false;
        }

    }

    public class CalculateID
    {
        private double nrmedHeight, nrmedIDNumber, curZ;
        // member
        //       private Vector2 _rightHandToHipCenter, _rightHandToShoulderCenter, _rightHandToElbowRight, _rightHandToHeadCenter;

        //        private bool _result = false;
        //        private bool _disable = true;

        public CalculateID()
        {
            //            _rightHandToHipCenter = new Vector2();
        }

        // process joint :: put it in "Update" routine
        public void process(SkeletonData skeleton)
        {
            Joint head = skeleton.Joints[JointID.Head].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint shoulder_center = skeleton.Joints[JointID.ShoulderCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint hip_center = skeleton.Joints[JointID.HipCenter].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint knee_left = skeleton.Joints[JointID.KneeLeft].ScaleTo(640, 480, 1.0f, 1.0f);
            Joint ankle_left = skeleton.Joints[JointID.AnkleLeft].ScaleTo(640, 480, 1.0f, 1.0f);

            double curHeight = 0;

            curHeight += calculateLength(head.Position.X, head.Position.Y, head.Position.Z, shoulder_center.Position.X, shoulder_center.Position.Y, shoulder_center.Position.Z);
            curHeight += calculateLength(shoulder_center.Position.X, shoulder_center.Position.Y, shoulder_center.Position.Z, hip_center.Position.X, hip_center.Position.Y, hip_center.Position.Z);
            curHeight += calculateLength(hip_center.Position.X, hip_center.Position.Y, hip_center.Position.Z, knee_left.Position.X, knee_left.Position.Y, knee_left.Position.Z);
            //curHeight += calculateLength(knee_left.Position.X, knee_left.Position.Y, knee_left.Position.Z, ankle_left.Position.X, ankle_left.Position.Y, ankle_left.Position.Z);

            curZ = hip_center.Position.Z;
            nrmedHeight = curHeight; // / curZ;


        }

        public double getNrmedHeight()
        {
            return nrmedHeight;
        }
        public double getNrmedIDNumber()
        {
            return nrmedIDNumber;
        }

        public double calculateLength(double x1, double y1, double z1, double x2, double y2, double z2)
        {
            double len;
            len = Math.Abs(x1 - x2) * Math.Abs(x1 - x2) + Math.Abs(y1 - y2) * Math.Abs(y1 - y2) + Math.Abs(z1 - z2) * Math.Abs(z1 - z2);
            return Math.Sqrt(len);

        }

        public double getCurZ()
        {
            return curZ;
        }

    }
}