﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Net;
using System.IO;
using System.Threading;
using Bespoke.Common;
using Bespoke.Common.Osc; // for osc transmitter


namespace hcXNA
{
    public class hcNetSender
    {
        // members
        private string _httpServer;
        private IPEndPoint _oscServer;
        private string _defaultOscPath;

        private bool _ACTIVE = true;
        private string _lastLogMsg = "";
 
        // ctor
        public hcNetSender(hcSetting setting)
        {
            // plain http server
            _httpServer = "http://" + setting.ServerURL + ":" + setting.HttpPort.ToString();

            // osc stuff
            OscPacket.LittleEndianByteOrder = false; // reverse the byte order
            _oscServer = new IPEndPoint(IPAddress.Parse(setting.OscServerURL), setting.OscPort);
            _defaultOscPath = setting.DefaultOscPath;

            // if local_mode switch is false
            if (setting.LOCAL_MODE) _ACTIVE = false;
        }

        // example: send Msg, path only
        public void sendOscMsg(string path)
        {
            if (!_ACTIVE) return; // only if active

            OscMessage msg = new OscMessage(_oscServer, _defaultOscPath + path);
            msg.Send(_oscServer);
        }

        // send event: a path with integer
        public void sendOscMsg(string path, int i)
        {
            if (!_ACTIVE) return; // only if active

            OscMessage msg = new OscMessage(_oscServer, _defaultOscPath + path);
            msg.Append(i);
            msg.Send(_oscServer);
        }        

        // send http message to Server
        public void sendHttpRequest(string msg)
        {
            if (!_ACTIVE) return; // only if active

            // try & catch
            HttpWebRequest myHttpWebRequest = (HttpWebRequest)WebRequest.Create(_httpServer + msg);
            WebResponse myResponse = myHttpWebRequest.GetResponse();
            myResponse.Close();
        }

        // polling & parsing
        public string pollServer()
        {
            if (!_ACTIVE) return "master,/photodesk"; // send bogus msg when local mode

            HttpWebRequest myHttpWebRequest = (HttpWebRequest)WebRequest.Create(_httpServer + "/poll");
            WebResponse myResponse = myHttpWebRequest.GetResponse();
            StreamReader reader = new StreamReader(myResponse.GetResponseStream(), Encoding.UTF8);
            string stringResponse = reader.ReadToEnd();
            myResponse.Close();
            return stringResponse;
        }

        // toggle
        public void toggle()
        {
            _ACTIVE = !_ACTIVE;
        }

        // get log message
        public string getLogMsg()
        {
            return _lastLogMsg;
        }
    }
}