﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.Graphics;

using Microsoft.Research.Kinect.Nui;
using Coding4Fun.Kinect.Wpf;

namespace hcXNA
{
    // public enum
    public enum hcState { IDLE, IDENTIFYING, IDENTIFIED, PHOTODESK, BROWSING, ZOOMIN, RATING };

    // stateClass :: hcSM (singleton)
    public class hcSM
    {
        // --------------------------------------------------------------- members
        // parameters
        hcState _currentState = hcState.IDLE;
        private TimeSpan _eTime = TimeSpan.Zero; // elapsed time
        private TimeSpan _timer = TimeSpan.Zero; // statemachine timer
        private TimeSpan _freezeOut = TimeSpan.Zero; // freeze out for swipes
        private TimeSpan _away =  TimeSpan.Zero; // away timer
        // private bool _LOOP = false;

        // state-specific params
        private int _lastRating = -1;
        private int _lastQuadrant = -1;

        // gester finders (in extended class)
        Proxim myProxim;
        Swiper mySwiper;
        Swiper2 mySwiper2;
        PassControl myPassControl;
        ZoomIn myZoomin; // added ver1
        DoubleSwing myDoubleSwing;
        GrabAndDrop myGrabAndDrop; // needs to be fixed
        StarRating myStarRating;
        twoHands my2h;        

        // netsender
        hcNetSender myNetSender;

        // logging
        hcConsole _console;

        // slave or master
        private string _clientID;
        private bool _isInControl; // this is master
        private TimeSpan _pollTimer = TimeSpan.Zero;

        // --------------------------------------------------------------- methods
        // ctor
        public hcSM(hcSetting setting, hcConsole console)
        {
            // init gestureFinders
            myProxim = new Proxim(setting);
            mySwiper = new Swiper();
            mySwiper2 = new Swiper2();
            myZoomin = new ZoomIn(); // added ver1 
            myPassControl = new PassControl(setting); 
            myDoubleSwing = new DoubleSwing(setting);
            myGrabAndDrop = new GrabAndDrop(setting);
            myStarRating = new StarRating(setting);
            my2h = new twoHands(setting);

            // netsender
            myNetSender = new hcNetSender(setting);

            // console
            _console = console;

            // control
            _clientID = setting.clientID;
            if (_clientID == "master") _isInControl = true;
            else _isInControl = false;

            // init the server
            sendNetMessage("/" + setting.clientID);
            sendNetMessage("/idle");
        }

        // update
        public void update(GameTime time)
        {
            // get elapsed time
            _eTime = time.ElapsedGameTime;

            // check control
            if (_isInControl)
            {
                // state switching
                switch (_currentState)
                {
                    case hcState.IDLE:
                        doIdle();
                        break;
                    case hcState.IDENTIFYING:
                        doIdentifying();
                        break;
                    case hcState.IDENTIFIED:
                        doIdentified();
                        break;
                    case hcState.PHOTODESK:
                        doPhotodesk();
                        break;
                    case hcState.BROWSING:
                        doBrowsing();
                        break;
                    case hcState.ZOOMIN:
                        doZoomin();
                        break;
                    case hcState.RATING:
                        doRating();
                        break;
                    default:
                        break;
                }

                // send realtime visual feedback on two hands
                sendTwohands(my2h.getString());
            }

            // only when this is slave; it's not in control
            else
            {
                // polling GUI server
                if (_pollTimer > TimeSpan.FromSeconds(1))
                {
                    _pollTimer = TimeSpan.Zero; // reset timer

                    // get polling msg from server
                    string pollResult = myNetSender.pollServer();
                    string[] words = pollResult.Split(',');
                    if (words[0] == _clientID)
                    {
                        _console.log(":::: " + _clientID + " in Control");
                        _isInControl = true;
                        return;
                    }

                    // process the url
                    // _console.log(words[0] + ":" + words[1]);
                    
                    string[] dest = words[1].Split('/');
                    switch (dest[1])
                    {
                        case "idle":
                            transit(hcState.IDLE);
                            break;
                        case "photodesk":
                            transit(hcState.PHOTODESK);
                            break;
                        case "browsing":
                            transit(hcState.BROWSING);
                            break;
                        case "zoomin":
                            transit(hcState.ZOOMIN);
                            break;
                        case "rating":
                            transit(hcState.RATING);
                            break;
                        default:
                            break;
                    }
                }
                else
                {
                    _pollTimer += _eTime; // advance time
                }
            }

            // end of func
            return;
        }

        // processSkeletons
        public void processSkeletons(SkeletonData sd)
        {
            // process skeleton by state, only when in control
            if (_isInControl)
            {
                switch (_currentState)
                {
                    case hcState.IDLE:
                        myProxim.processSkeleton(sd);
                        break;
                    case hcState.IDENTIFYING:
                        myProxim.processSkeleton(sd);
                        break;
                    case hcState.IDENTIFIED:
                        myProxim.processSkeleton(sd);
                        break;
                    case hcState.PHOTODESK:
                        myProxim.processSkeleton(sd);
                        myGrabAndDrop.processSkeleton(sd);
                        mySwiper.processSkeleton(sd);
                        myPassControl.processSkeleton(sd);
                        break;
                    case hcState.BROWSING:
                        myProxim.processSkeleton(sd);
                        myStarRating.processSkeleton(sd);
                        mySwiper2.processSkeleton(sd);
                        myZoomin.processSkeleton(sd);
                        myDoubleSwing.processSkeleton(sd);
                        myPassControl.processSkeleton(sd);
                        break;
                    case hcState.ZOOMIN:
                        myProxim.processSkeleton(sd);
                        mySwiper2.processSkeleton(sd);
                        myDoubleSwing.processSkeleton(sd);
                        myPassControl.processSkeleton(sd);
                        break;
                    case hcState.RATING:
                        myProxim.processSkeleton(sd);
                        myDoubleSwing.processSkeleton(sd);
                        myStarRating.processSkeleton(sd);
                        break;
                    default:
                        break;
                }

                // update two hands
                my2h.processSkeleton(sd);
            }            
        }


        // ---------------------------------------------------------------
        // doIdle
        private void doIdle()
        {
            // if user is in sweetspot - add up time, else reset it
            if (myProxim.getDiscrete() > 1) _timer += _eTime;
            else _timer = TimeSpan.Zero;

            // if timer passed 1 sec, initiate identifying mode
            if (_timer > TimeSpan.FromSeconds(1))
            {
                _away = TimeSpan.FromSeconds(3); // charge up 3 sec
                // sendNetMessage("/usericon/gray");
                sendNetMessage("/identifying");
                transit(hcState.IDENTIFYING);
            }
        }

        // doIdentifying
        private void doIdentifying()
        {
            // check user presence
            if (isUserAway()) return;

            // advance time
            _timer += _eTime;

            // if timer passed 1 sec, initiate identifying mode
            if (_timer > TimeSpan.FromSeconds(1))
            {
                // int id = identify.getDiscrete();
                int id = 1;
                // sendNetMessage("/usericon/green");
                sendNetMessage("/identified", id);
                transit(hcState.IDENTIFIED);
            }
        }

        // doIdentified
        private void doIdentified()
        {
            // check user presence
            if (isUserAway()) return;

            // advance time
            _timer += _eTime;

            // if timer passed 2 sec, initiate photodesk mode
            if (_timer > TimeSpan.FromSeconds(1))
            {
                sendNetMessage("/photodesk");
                transit(hcState.PHOTODESK);
                _freezeOut = TimeSpan.FromSeconds(2); // freeze out for 2 sec
            }
        }

        // doPhotodesk
        private void doPhotodesk()
        {
            // if user is away for 3 seconds, go to idle
            if (isUserAway()) return;

            // check gestures with freeze out
            if (_freezeOut < TimeSpan.Zero)
            {
                // check : hovering quadrants
                int quad = myGrabAndDrop.getDiscrete();
                if (quad > 0 && quad < 5)
                {
                    sendNetMessage("/photodesk/hover", quad);
                    _freezeOut = TimeSpan.FromSeconds(0.5);
                }

                // check : grabbing a quadrant
                if (quad > 0 && quad < 5 && myGrabAndDrop.getDecision())
                {
                    sendNetMessage("/browsing", quad); // send quadrant
                    _lastQuadrant = quad;
                    transit(hcState.BROWSING);
                    _freezeOut = TimeSpan.FromSeconds(2); // freeze out for 2 sec
                }

                // check: left/right swipe
                int amt = mySwiper.getResult();
                if (amt < 0)
                {
                    sendNetMessage("/photodesk/scrollleft");
                    _freezeOut = TimeSpan.FromSeconds(0.75);
                }
                else if (amt > 0)
                {
                    sendNetMessage("/photodesk/scrollright");
                    _freezeOut = TimeSpan.FromSeconds(0.75);
                }

                // check passControl
                if (myPassControl.getDecision())
                {
                    sendNetMessage("/yield");
                    _freezeOut = TimeSpan.FromSeconds(1);
                    _isInControl = false;
                }
            }
            else _freezeOut -= _eTime;
        }

        // doBrowsing
        private void doBrowsing()
        {
            // if user is away for 3 seconds, go to idle
            if (isUserAway()) return;

            // check gestures with freeze out
            if (_freezeOut < TimeSpan.Zero)
            {
                // if doubleswing, go back to photodesk
                if (myDoubleSwing.getDecision())
                {
                    sendNetMessage("/photodesk");
                    transit(hcState.PHOTODESK);
                    _freezeOut = TimeSpan.FromSeconds(2); // freeze out for 2 sec
                }

                // check left, right swipe
                int amt = mySwiper2.getResult();
                if (amt < 0)
                {
                    sendNetMessage("/browsing/scrollleft", -amt);
                    _freezeOut = TimeSpan.FromSeconds(0.75);
                }
                else if (amt > 0)
                {
                    sendNetMessage("/browsing/scrollright", amt);
                    _freezeOut = TimeSpan.FromSeconds(0.75);
                }

                // check grab-n-rate: if hands are together, go to rating
                if (myStarRating.isRatingStarted())
                {
                    _lastRating = -1; // init rating var
                    sendNetMessage("/rating");
                    _freezeOut = TimeSpan.FromSeconds(1);
                    transit(hcState.RATING);
                }

                // check passControl
                if (myPassControl.getDecision())
                {
                    sendNetMessage("/yield");
                    _freezeOut = TimeSpan.FromSeconds(1);
                    _isInControl = false;
                }

                // check zoomin
                if (myZoomin.getResult())
                {
                    sendNetMessage("/zoomin");
                    transit(hcState.ZOOMIN);
                    _freezeOut = TimeSpan.FromSeconds(1);
                }
            }
            else _freezeOut -= _eTime;            
        }

        // doZoomin
        private void doZoomin()
        {
            // if user is away for 3 seconds, go to idle
            if (isUserAway()) return;

            // check gestures with freeze out
            if (_freezeOut < TimeSpan.Zero)
            {
                // if doubleswing, go back to zoomout browsing
                if (myDoubleSwing.getDecision())
                {
                    sendNetMessage("/zoomout");
                    transit(hcState.BROWSING);
                    _freezeOut = TimeSpan.FromSeconds(2); // freeze out for 2 sec
                }

                // check left, right swipe
                int amt = mySwiper2.getResult();
                if (amt < 0)
                {
                    sendNetMessage("/zoomin/scrollleft");
                    _freezeOut = TimeSpan.FromSeconds(0.75);
                }
                else if (amt > 0)
                {
                    sendNetMessage("/zoomin/scrollright");
                    _freezeOut = TimeSpan.FromSeconds(0.75);
                }

                // check passControl
                if (myPassControl.getDecision())
                {
                    sendNetMessage("/yield");
                    _freezeOut = TimeSpan.FromSeconds(1);
                    _isInControl = false;
                }
            }
            else _freezeOut -= _eTime;
        }

        // doRating
        private void doRating()
        {
            // if user is away for 3 seconds, go to idle
            if (isUserAway()) return;

            // check gestures with freeze out
            /*
            if (_freezeOut < TimeSpan.Zero)
            {
                // if doubleswing, go back to browsing
                if (myDoubleSwing.getDecision())
                {
                    // sendNetMessage("/browsing", _lastQuadrant);
                    sendNetMessage("/rating/finish");
                    transit(hcState.BROWSING);
                    _freezeOut = TimeSpan.FromSeconds(2);
                }
            }
            else _freezeOut -= _eTime;
            */

            // getting rate from the gesture finder
            int currentRating = myStarRating.getNumStar();
            if (_lastRating != currentRating)
            {
                if (currentRating > -1)
                {
                    sendNetMessage("/rating", currentRating);
                }
                _lastRating = currentRating;
            }

            // if rating confirmed
            if (myStarRating.isRatingConfirmed())
            {
                sendNetMessage("/rating/finish");
                transit(hcState.BROWSING);
                _freezeOut = TimeSpan.FromSeconds(1);
            }

            // go back to browsing, if you do something else
            else if (myStarRating.isRatingStarted() == false)
            {
                // sendNetMessage("/browsing", _lastQuadrant);
                sendNetMessage("/rating/cancel");
                transit(hcState.BROWSING);
                _freezeOut = TimeSpan.FromSeconds(1);
            }
        }

        // isUserAway(): if user is away for 3 seconds, go to Idle
        private bool isUserAway()
        {
            // check proxim
            if (myProxim.getDiscrete() > 0) _away = TimeSpan.FromSeconds(3);
            else _away -= _eTime;

            // if away is less than zero, go to Idle and return true (user is away)
            if (_away < TimeSpan.Zero)
            {
                _away = TimeSpan.Zero;
                // sendNetMessage("/usericon/none");
                sendNetMessage("/idle");
                transit(hcState.IDLE);

                return true;
            }

            // else return false (user is present)
            return false;
        }

        // transit to next state
        public void transit(hcState nextState)
        {
            _currentState = nextState;
            _timer = TimeSpan.Zero;
            // _LOOP = false;
        }

        // sendNetMessage(): 1 string
        public void sendNetMessage(string str)
        {
            // send http, osc message
            myNetSender.sendHttpRequest(str);
            myNetSender.sendOscMsg(str);

            // update log
            _console.log(str);
            
            // return
            return;
        }

        // sendNetMessage(): 1 string + 1 int
        public void sendNetMessage(string str, int i)
        {
            // send http, osc message
            myNetSender.sendHttpRequest(str + "/" + i.ToString());
            myNetSender.sendOscMsg(str, i);
            
            // update log
            _console.log(str + "/" + i.ToString());
            
            // return
            return;
        }

        // send realtime visual feedback
        public void sendTwohands(string str)
        {
            // send http
            myNetSender.sendHttpRequest("/handposition" + str);

            // return
            return;
        }

        // ------------------------------------------------------------- getter
        // getCurrentState
        public hcState getState()
        {
            return _currentState;
        }

        // getTimer
        public int getTimer()
        {
            return _timer.Seconds;
        }

        // getAway
        public int getAway()
        {
            return _away.Seconds;
        }

        // getFreezeOut
        public int getFreezeOut()
        {
            return _freezeOut.Seconds;
        }
    }
}
